/*************************************************************************\
*                  Copyright (C) Michael Kerrisk, 2014.                   *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Affero General Public License as published   *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
* See the file COPYING.agpl-v3 for details.                               *
\*************************************************************************/

/* Supplementary program for Chapter 17 */

  1 #include <sys/acl.h>
  2 #include "ugid_functions.h"
  3 #include "tlpi_hdr.h"
  4
  5 #define MAX_ENTRIES 10000   /* Maximum entries that we can handle in an ACL */
  6
  7 struct AccessControlEntry {     /* Represent a single ACL entry */
  8     acl_tag_t   tag;            /* Tag type */
  9     id_t        qual;           /* Optional tag qualifier (UID or GID) */
 10     int         perms;          /* Permissions bit mask */
 11 };
 12
 13 static void
 14 usageError(char *progName, char *msg, Boolean shortUsage)
 15 {
 16     if (msg != NULL)
 17         fprintf(stderr, "%s\n", msg);
 18
 19     if (shortUsage) {
 20         fprintf(stderr, "Type '%s --help' for usage information\n", progName);
 21         exit(EXIT_FAILURE);
 22     }
 23
 24     fprintf(stderr, "Usage: %s -m acl [-d] [-n] file...\n", progName);
 25     fprintf(stderr, "   or: %s -x acl [-d] [-n] file...\n", progName);
 26     fprintf(stderr, "   or: %s -k file...\n\n", progName);
 27     fprintf(stderr, "   or: %s -V acl\n\n", progName);
 28 #define fpe(msg) fprintf(stderr, "      " msg);
 29     fpe("-m  Modify/create ACL entries\n");
 30     fpe("-x  Remove ACL entries\n");
 31     fpe("-k  Remove default ACL\n\n");
 32     fpe("-V  Check validity of 'acl'\n\n");
 33     fpe("'acl' consists of one or more comma-separated entries of the form:\n");
 34     fpe("\n            tag:[qualifier][:[perms]]\n\n");
 35     fpe("      'perms' are specified only for -m\n\n");
 36     fpe("-d  Apply operation to default ACL\n");
 37     fpe("-n  Don't recalculate mask entry\n");
 38     fpe("      NOTE: if you specify this option and have specified\n");
 39     fpe("      the -m option, then you may encounter errors if the\n");
 40     fpe("      file does not already have a mask entry.\n");
 41     exit(EXIT_FAILURE);
 42 }
 43
 44 /* Parse an ACL entry specification (the null-terminated string
 45    provided in 'entryStr') of the form:
 46
 47         tag:[qualifier][:[permissions]]
 48
 49    returning the parsed information in the structure 'ace'.
 50
 51    If 'permsReqd' is set, then the ACL entries must contain
 52    permission specifications (i.e., a colon followed by at least
 53    one of [-rwx], otherwise they must not.
 54
 55    Return TRUE if the specification parsed okay, or FALSE otherwise. */
 56
 57 static Boolean
 58 parseEntrySpec(char *entryStr, struct AccessControlEntry *ace,
 59                Boolean permsReqd)
 60 {
 61     char *colon1, *colon2;
 62     Boolean hasQual;            /* Is optional qualifier present? */
 63     Boolean hasPerms;           /* Are permissions specified? */
 64
 65     colon1 = strchr(entryStr, ':');
 66     if (colon1 == NULL) {
 67         fprintf(stderr, "Missing initial colon in ACL entry: %s\n", entryStr);
 68         return FALSE;
 69     }
 70
 71     hasQual = *(colon1 + 1) != '\0' && *(colon1 + 1) != ':';
 72
 73     *colon1 = '\0';     /* Add terminator to tag type */
 74
 75     colon2 = strchr(colon1 + 1, ':');
 76     hasPerms = colon2 != NULL && *(colon2 + 1) != '\0';
 77
 78     if (hasPerms && !permsReqd) {
 79         fprintf(stderr, "Cannot specify permissions here\n");
 80         return FALSE;
 81     }
 82
 83     if (!hasPerms && permsReqd) {
 84         fprintf(stderr, "Must specify permissions\n");
 85         return FALSE;
 86     }
 87
 88     /* Determine tag type, depending on tag string and presence
 89        of qualifier after the first ':' */
 90
 91     if (strcmp(entryStr, "u") == 0 || strcmp(entryStr, "user") == 0)
 92         ace->tag = hasQual ? ACL_USER : ACL_USER_OBJ;
 93     else if (strcmp(entryStr, "g") == 0 || strcmp(entryStr, "group") == 0)
 94         ace->tag = hasQual ? ACL_GROUP : ACL_GROUP_OBJ;
 95     else if (strcmp(entryStr, "o") == 0 || strcmp(entryStr, "other") == 0)
 96         ace->tag = ACL_OTHER;
 97     else if (strcmp(entryStr, "m") == 0 || strcmp(entryStr, "mask") == 0)
 98         ace->tag = ACL_MASK;
 99     else {
100         fprintf(stderr, "Bad tag: %s\n", entryStr);
101         return FALSE;
102     }
103
104     /* For ACL_USER and ACL_GROUP tags, extract a UID / GID from qualifier */
105
106     if (colon2 != NULL)
107         *colon2 = '\0';         /* Add terminator to qualifier */
108
109     ace->qual = 0;
110
111     if (ace->tag == ACL_USER) {
112         ace->qual = userIdFromName(colon1 + 1);
113         if (ace->qual == -1) {
114             fprintf(stderr, "Bad user ID: %s\n", colon1 + 1);
115             return FALSE;
116         }
117     } else if (ace->tag == ACL_GROUP) {
118         ace->qual = groupIdFromName(colon1 + 1);
119         if (ace->qual == -1) {
120             fprintf(stderr, "Bad group ID: %s\n", colon1 + 1);
121             return FALSE;
122         }
123     }
124
125     /* If a permissions string was present, return it as a bit mask */
126
127     if (hasPerms) {
128         char *p;
129
130         ace->perms = 0;
131
132         /* We're not too thorough here -- we don't check for multiple
133            instances of [-rwx], or check if the permissions string is
134            longer than three characters... */
135
136         for (p = colon2 + 1; *p != '\0'; p++) {
137             if (*p == 'r')
138                 ace->perms |= ACL_READ;
139             else if (*p == 'w')
140                 ace->perms |= ACL_WRITE;
141             else if (*p == 'x')
142                 ace->perms |= ACL_EXECUTE;
143             else if (*p != '-') {
144                 fprintf(stderr, "Bad character in permissions "
145                         "string: %c\n", *p);
146                 return FALSE;
147             }
148         }
149     }
150
151     return TRUE;
152 }
153
154 /* Parse a text form ACL, returning information about the entries in
155    'aclist'. On success, return the number of ACL entries found; on
156    error return -1. */
157
158 static int
159 parseACL(char *aclStr, struct AccessControlEntry aclist[],
160         Boolean permsReqd)
161 {
162     char *nextEntry, *comma;
163     int n;
164
165     n = 0;
166     for (nextEntry = aclStr; ; nextEntry = comma + 1) {
167
168         if (n >= MAX_ENTRIES) {
169             fprintf(stderr, "Too many entries in ACL\n");
170             return -1;
171         }
172
173         comma = strchr(nextEntry, ',');
174         if (comma != NULL)
175             *comma = '\0';
176
177         if (!parseEntrySpec(nextEntry, &aclist[n], permsReqd))
178             return -1;
179
180         if (comma == NULL)              /* This was the last entry */
181             break;
182
183         n++;
184     }
185
186     return n + 1;
187 }
188
189 /* Find the the ACL entry in 'acl' corresponding to the tag type and
190    qualifier in 'tag' and 'id'. Return the matching entry, or NULL
191    if no entry was found. */
192
193 static acl_entry_t
194 findEntry(acl_t acl, acl_tag_t tag, id_t qaul)
195 {
196     acl_entry_t entry;
197     acl_tag_t entryTag;
198     uid_t *uidp;
199     gid_t *gidp;
200     int ent, s;
201
202     for (ent = ACL_FIRST_ENTRY; ; ent = ACL_NEXT_ENTRY) {
203         s = acl_get_entry(acl, ent, &entry);
204         if (s == -1)
205             errExit("acl_get_entry");
206
207         if (s == 0)
208             return NULL;
209
210         if (acl_get_tag_type(entry, &entryTag) == -1)
211             errExit("acl_get_tag_type");
212
213         if (tag == entryTag) {
214             if (tag == ACL_USER) {
215                 uidp = acl_get_qualifier(entry);
216                 if (uidp == NULL)
217                     errExit("acl_get_qualifier");
218
219                 if (qaul == *uidp) {
220                     if (acl_free(uidp) == -1)
221                         errExit("acl_free");
222                     return entry;
223                 } else {
224                     if (acl_free(uidp) == -1)
225                         errExit("acl_free");
226                 }
227
228             } else if (tag == ACL_GROUP) {
229                 gidp = acl_get_qualifier(entry);
230                 if (gidp == NULL)
231                     errExit("acl_get_qualifier");
232
233                 if (qaul == *gidp) {
234                     if (acl_free(gidp) == -1)
235                         errExit("acl_free");
236                     return entry;
237                 } else {
238                     if (acl_free(gidp) == -1)
239                         errExit("acl_free");
240                 }
241
242             } else {
243                 return entry;
244             }
245         }
246     }
247 }
248
249 /* Set the permissions in 'perms' in the ACL entry 'entry' */
250
251 static void
252 setPerms(acl_entry_t entry, int perms)
253 {
254     acl_permset_t permset;
255
256     if (acl_get_permset(entry, &permset) == -1)
257         errExit("acl_get_permset");
258
259     if (acl_clear_perms(permset) == -1)
260         errExit("acl_clear_perms");
261
262     if (perms & ACL_READ)
263         if (acl_add_perm(permset, ACL_READ) == -1)
264             errExit("acl_add_perm");
265     if (perms & ACL_WRITE)
266         if (acl_add_perm(permset, ACL_WRITE) == -1)
267             errExit("acl_add_perm");
268     if (perms & ACL_EXECUTE)
269         if (acl_add_perm(permset, ACL_EXECUTE) == -1)
270             errExit("acl_add_perm");
271
272     if (acl_set_permset(entry, permset) == -1)
273         errExit("acl_set_permset");
274 }
275
276 int
277 main(int argc, char *argv[])
278 {
279     Boolean recalcMask, useDefaultACL;
280     Boolean modifyACL, removeACL, removeDefaultACL, checkValidity;
281     int optCnt, j, opt, numEntries, en;
282     acl_type_t type;
283     char *aclSpec;
284     acl_t acl;
285     acl_entry_t entry;
286     struct AccessControlEntry aclist[MAX_ENTRIES];
287
288     if (argc < 2 || strcmp(argv[1], "--help") == 0)
289         usageError(argv[0], NULL, FALSE);
290
291     /* Parse command-line options */
292
293     recalcMask = TRUE;
294     useDefaultACL = FALSE;
295     modifyACL = FALSE;
296     removeACL = FALSE;
297     checkValidity = FALSE;
298     removeDefaultACL = FALSE;
299     optCnt = 0;
300
301     while ((opt = getopt(argc, argv, "m:x:kdnV:")) != -1) {
302         switch (opt) {
303         case 'm':
304             modifyACL = TRUE;
305             aclSpec = optarg;
306             optCnt++;
307             break;
308
309         case 'x':
310             removeACL = TRUE;
311             aclSpec = optarg;
312             optCnt++;
313             break;
314
315         case 'k':
316             removeDefaultACL = TRUE;
317             optCnt++;
318             break;
319
320         case 'V':
321             checkValidity = TRUE;
322             aclSpec = optarg;
323             optCnt++;
324             break;
325
326         case 'd':
327             useDefaultACL = TRUE;
328             break;
329
330         case 'n':
331             recalcMask = FALSE;
332             break;
333
334         default:
335             usageError(argv[0], "Bad option\n", TRUE);
336             break;
337         }
338     }
339
340     if (optCnt != 1)
341         usageError(argv[0], "Specify exactly one of -m, -x, -k, or -V\n", TRUE);
342
343     if (checkValidity && useDefaultACL)
344         usageError(argv[0], "Can't specify -d with -V\n", TRUE);
345
346     if (checkValidity) {
347         if (parseACL(aclSpec, aclist, TRUE) == -1) {
348             fatal("Bad ACL entry specification");
349         } else {
350             printf("ACL is valid\n");
351             exit(EXIT_SUCCESS);
352         }
353     }
354
355     if (modifyACL || removeACL) {
356         numEntries = parseACL(aclSpec, aclist, modifyACL);
357         if (numEntries == -1)
358             usageError(argv[0], "Bad ACL specification\n", TRUE);
359     }
360
361     type = useDefaultACL ? ACL_TYPE_DEFAULT : ACL_TYPE_ACCESS;
362
363     /* Perform the operation on each file argument */
364
365     for (j = optind; j < argc; j++) {
366         if (removeDefaultACL) {
367             if (acl_delete_def_file(argv[j]) == -1)
368                 errExit("acl_delete_def_file: %s", argv[j]);
369
370         } else if (modifyACL || removeACL) {
371
372             acl = acl_get_file(argv[j], type);
373             if (acl == NULL)
374                 errExit("acl_get_file");
375
376             /* Apply each of the entries in 'aclist' to the
377                current file */
378
379             for (en = 0; en < numEntries; en++) {
380                 entry = findEntry(acl, aclist[en].tag, aclist[en].qual);
381
382                 if (removeACL) {
383                     if (entry != NULL)
384                         if (acl_delete_entry(acl, entry) == -1)
385                             errExit("acl_delete_entry");
386
387                 } else {        /* modifyACL */
388
389                     if (entry == NULL) {
390
391                         /* Entry didn't exist in ACL -- create a new
392                            entry with required tag and qualifier */
393
394                         if (acl_create_entry(&acl, &entry) == -1)
395                             errExit("acl_create_entry");
396                         if (acl_set_tag_type(entry, aclist[en].tag) == -1)
397                             errExit("acl_set_tag_type");
398                         if (aclist[en].tag == ACL_USER ||
399                                 aclist[en].tag == ACL_GROUP)
400                             if (acl_set_qualifier(entry,
401                                         &aclist[en].qual) == -1)
402                                 errExit("acl_set_qualifier");
403                     }
404
405                     setPerms(entry, aclist[en].perms);
406                 }
407
408                 /* Recalculate the mask entry if requested */
409
410                 if (recalcMask)
411                     if (acl_calc_mask(&acl) == -1)
412                         errExit("acl_calc_mask");
413
414                 /* Update the file ACL */
415
416                 if (acl_valid(acl) == -1)
417                     errExit("acl_valid");
418
419                 if (acl_set_file(argv[j], type, acl) == -1)
420                     errExit("acl_set_file");
421             }
422
423             if (acl_free(acl) == -1)
424                 errExit("acl_free");
425         } else {
426             fatal("Bad logic!");
427         }
428     }
429
430     exit(EXIT_SUCCESS);
431 }
