/*************************************************************************\
*                  Copyright (C) Michael Kerrisk, 2014.                   *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Affero General Public License as published   *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
* See the file COPYING.agpl-v3 for details.                               *
\*************************************************************************/

/* Supplementary program for Chapter 28 */

  1 #define _GNU_SOURCE
  2 #include <string.h>
  3 #include <signal.h>
  4 #include <sys/stat.h>
  5 #include <sys/wait.h>
  6 #include <fcntl.h>
  7 #include <sched.h>
  8 #include "print_wait_status.h"
  9 #include "tlpi_hdr.h"
 10
 11 #ifndef CHILD_SIG
 12 #define CHILD_SIG SIGUSR1       /* Signal to be generated on termination
 13                                    of cloned child */
 14 #endif
 15
 16 typedef struct {        /* For passing info to child startup function */
 17     int    fd;
 18     mode_t umask;
 19     int    exitStatus;
 20     int    signal;
 21 } ChildParams;
 22
 23 static int              /* Startup function for cloned child */
 24 childFunc(void *arg)
 25 {
 26     ChildParams *cp;
 27
 28     printf("Child:  PID=%ld PPID=%ld\n", (long) getpid(), (long) getppid());
 29
 30     cp = (ChildParams *) arg;   /* Cast arg to true form */
 31
 32     /* The following changes may affect parent */
 33
 34     umask(cp->umask);
 35     if (close(cp->fd) == -1)
 36         errExit("child:close");
 37     if (signal(cp->signal, SIG_DFL) == SIG_ERR)
 38         errExit("child:signal");
 39
 40     return cp->exitStatus;      /* Child terminates now */
 41 }
 42
 43 static void             /* Handler for child termination signal */
 44 grimReaper(int sig)
 45 {
 46     int savedErrno;
 47
 48     /* UNSAFE: This handler uses non-async-signal-safe functions
 49        (printf(), strsignal(); see Section 21.1.2) */
 50
 51     savedErrno = errno;         /* In case we change 'errno' here */
 52
 53     printf("Caught signal %d (%s)\n", sig, strsignal(sig));
 54
 55     errno = savedErrno;
 56 }
 57
 58 static void
 59 usageError(char *progName)
 60 {
 61     fprintf(stderr, "Usage: %s [arg]\n", progName);
 62 #define fpe(str) fprintf(stderr, "        " str)
 63     fpe("'arg' can contain the following letters:\n");
 64     fpe("    d - share file descriptors (CLONE_FILES)\n");
 65     fpe("    f - share file-system information (CLONE_FS)\n");
 66     fpe("    s - share signal dispositions (CLONE_SIGHAND)\n");
 67     fpe("    v - share virtual memory (CLONE_VM)\n");
 68     exit(EXIT_FAILURE);
 69 }
 70
 71 int
 72 main(int argc, char *argv[])
 73 {
 74     const int STACK_SIZE = 65536;       /* Stack size for cloned child */
 75     char *stack;                        /* Start of stack buffer area */
 76     char *stackTop;                     /* End of stack buffer area */
 77     int flags;                          /* Flags for cloning child */
 78     ChildParams cp;                     /* Passed to child function */
 79     const mode_t START_UMASK = S_IWOTH; /* Initial umask setting */
 80     struct sigaction sa;
 81     char *p;
 82     int status;
 83     ssize_t s;
 84     pid_t pid;
 85
 86     printf("Parent: PID=%ld PPID=%ld\n", (long) getpid(), (long) getppid());
 87
 88     /* Set up an argument structure to be passed to cloned child, and
 89        set some process attributes that will be modified by child */
 90
 91     cp.exitStatus = 22;                 /* Child will exit with this status */
 92
 93     umask(START_UMASK);                 /* Initialize umask to some value */
 94     cp.umask = S_IWGRP;                 /* Child sets umask to this value */
 95
 96     cp.fd = open("/dev/null", O_RDWR);  /* Child will close this fd */
 97     if (cp.fd == -1)
 98         errExit("open");
 99
100     cp.signal = SIGTERM;                /* Child will change disposition */
101     if (signal(cp.signal, SIG_IGN) == SIG_ERR)
102         errExit("signal");
103
104     /* Initialize clone flags using command-line argument (if supplied) */
105
106     flags = 0;
107     if (argc > 1) {
108         for (p = argv[1]; *p != '\0'; p++) {
109             if      (*p == 'd') flags |= CLONE_FILES;
110             else if (*p == 'f') flags |= CLONE_FS;
111             else if (*p == 's') flags |= CLONE_SIGHAND;
112             else if (*p == 'v') flags |= CLONE_VM;
113             else usageError(argv[0]);
114         }
115     }
116
117     /* Allocate stack for child */
118
119     stack = malloc(STACK_SIZE);
120     if (stack == NULL)
121         errExit("malloc");
122     stackTop = stack + STACK_SIZE;  /* Assume stack grows downward */
123
124     /* Establish handler to catch child termination signal */
125
126     if (CHILD_SIG != 0) {
127         sigemptyset(&sa.sa_mask);
128         sa.sa_flags = SA_RESTART;
129         sa.sa_handler = grimReaper;
130         if (sigaction(CHILD_SIG, &sa, NULL) == -1)
131             errExit("sigaction");
132     }
133
134     /* Create child; child commences execution in childFunc() */
135
136     if (clone(childFunc, stackTop, flags | CHILD_SIG, &cp) == -1)
137         errExit("clone");
138
139     /* Parent falls through to here. Wait for child; __WCLONE option is
140        required for child notifying with signal other than SIGCHLD. */
141
142     pid = waitpid(-1, &status, (CHILD_SIG != SIGCHLD) ? __WCLONE : 0);
143     if (pid == -1)
144         errExit("waitpid");
145
146     printf("    Child PID=%ld\n", (long) pid);
147     printWaitStatus("    Status: ", status);
148
149     /* Check whether changes made by cloned child have affected parent */
150
151     printf("Parent - checking process attributes:\n");
152     if (umask(0) != START_UMASK)
153         printf("    umask has changed\n");
154     else
155         printf("    umask has not changed\n");
156
157     s = write(cp.fd, "Hello world\n", 12);
158     if (s == -1 && errno == EBADF)
159         printf("    file descriptor %d has been closed\n", cp.fd);
160     else if (s == -1)
161         printf("    write() on file descriptor %d failed (%s)\n",
162                 cp.fd, strerror(errno));
163     else
164         printf("    write() on file descriptor %d succeeded\n", cp.fd);
165
166     if (sigaction(cp.signal, NULL, &sa) == -1)
167         errExit("sigaction");
168     if (sa.sa_handler != SIG_IGN)
169         printf("    signal disposition has changed\n");
170     else
171         printf("    signal disposition has not changed\n");
172
173     exit(EXIT_SUCCESS);
174 }
