/*************************************************************************\
*                  Copyright (C) Michael Kerrisk, 2014.                   *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Affero General Public License as published   *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
* See the file COPYING.agpl-v3 for details.                               *
\*************************************************************************/

/* Listing 47-8 */

  1 #include <sys/types.h>
  2 #include <sys/sem.h>
  3 #include <ctype.h>
  4 #include "curr_time.h"          /* Declaration of currTime() */
  5 #include "tlpi_hdr.h"
  6
  7 #define MAX_SEMOPS 1000         /* Maximum operations that we permit for
  8                                    a single semop() */
  9
 10 static void
 11 usageError(const char *progName)
 12 {
 13     fprintf(stderr, "Usage: %s semid op[,op...] ...\n\n", progName);
 14     fprintf(stderr, "'op' is either: <sem#>{+|-}<value>[n][u]\n");
 15     fprintf(stderr, "            or: <sem#>=0[n]\n");
 16     fprintf(stderr, "       \"n\" means include IPC_NOWAIT in 'op'\n");
 17     fprintf(stderr, "       \"u\" means include SEM_UNDO in 'op'\n\n");
 18     fprintf(stderr, "The operations in each argument are "
 19                     "performed in a single semop() call\n\n");
 20     fprintf(stderr, "e.g.: %s 12345 0+1,1-2un\n", progName);
 21     fprintf(stderr, "      %s 12345 0=0n 1+1,2-1u 1=0\n", progName);
 22     exit(EXIT_FAILURE);
 23 }
 24
 25 /* Parse comma-delimited operations in 'arg', returning them in the
 26    array 'sops'. Return number of operations as function result. */
 27
 28 static int
 29 parseOps(char *arg, struct sembuf sops[])
 30 {
 31     char *comma, *sign, *remaining, *flags;
 32     int numOps;                 /* Number of operations in 'arg' */
 33
 34     for (numOps = 0, remaining = arg; ; numOps++) {
 35         if (numOps >= MAX_SEMOPS)
 36             cmdLineErr("Too many operations (maximum=%d): \"%s\"\n",
 37                         MAX_SEMOPS, arg);
 38
 39         if (*remaining == '\0')
 40             fatal("Trailing comma or empty argument: \"%s\"", arg);
 41         if (!isdigit((unsigned char) *remaining))
 42             cmdLineErr("Expected initial digit: \"%s\"\n", arg);
 43
 44         sops[numOps].sem_num = strtol(remaining, &sign, 10);
 45
 46         if (*sign == '\0' || strchr("+-=", *sign) == NULL)
 47             cmdLineErr("Expected '+', '-', or '=' in \"%s\"\n", arg);
 48         if (!isdigit((unsigned char) *(sign + 1)))
 49             cmdLineErr("Expected digit after '%c' in \"%s\"\n", *sign, arg);
 50
 51         sops[numOps].sem_op = strtol(sign + 1, &flags, 10);
 52
 53         if (*sign == '-')                       /* Reverse sign of operation */
 54             sops[numOps].sem_op = - sops[numOps].sem_op;
 55         else if (*sign == '=')                  /* Should be '=0' */
 56             if (sops[numOps].sem_op != 0)
 57                 cmdLineErr("Expected \"=0\" in \"%s\"\n", arg);
 58
 59         sops[numOps].sem_flg = 0;
 60         for (;; flags++) {
 61             if (*flags == 'n')
 62                 sops[numOps].sem_flg |= IPC_NOWAIT;
 63             else if (*flags == 'u')
 64                 sops[numOps].sem_flg |= SEM_UNDO;
 65             else
 66                 break;
 67         }
 68
 69         if (*flags != ',' && *flags != '\0')
 70             cmdLineErr("Bad trailing character (%c) in \"%s\"\n", *flags, arg);
 71
 72         comma = strchr(remaining, ',');
 73         if (comma == NULL)
 74             break;                              /* No comma --> no more ops */
 75         else
 76             remaining = comma + 1;
 77     }
 78
 79     return numOps + 1;
 80 }
 81
 82 int
 83 main(int argc, char *argv[])
 84 {
 85     struct sembuf sops[MAX_SEMOPS];
 86     int ind, nsops;
 87
 88     if (argc < 2 || strcmp(argv[1], "--help") == 0)
 89         usageError(argv[0]);
 90
 91     for (ind = 2; argv[ind] != NULL; ind++) {
 92         nsops = parseOps(argv[ind], sops);
 93
 94         printf("%5ld, %s: about to semop()  [%s]\n", (long) getpid(),
 95                 currTime("%T"), argv[ind]);
 96
 97         if (semop(getInt(argv[1], 0, "semid"), sops, nsops) == -1)
 98             errExit("semop (PID=%ld)", (long) getpid());
 99
100         printf("%5ld, %s: semop() completed [%s]\n", (long) getpid(),
101                 currTime("%T"), argv[ind]);
102     }
103
104     exit(EXIT_SUCCESS);
105 }
