/*************************************************************************\
*                  Copyright (C) Michael Kerrisk, 2014.                   *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Affero General Public License as published   *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
* See the file COPYING.agpl-v3 for details.                               *
\*************************************************************************/

/* Listing 62-4 */

  1
  2 #include <termios.h>
  3 #include <signal.h>
  4 #include <ctype.h>
  5 #include "tty_functions.h"              /* Declarations of ttySetCbreak()
  6                                            and ttySetRaw() */
  7 #include "tlpi_hdr.h"
  8
  9 static struct termios userTermios;
 10                         /* Terminal settings as defined by user */
 11
 12 static void             /* General handler: restore tty settings and exit */
 13 handler(int sig)
 14 {
 15     if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &userTermios) == -1)
 16         errExit("tcsetattr");
 17     _exit(EXIT_SUCCESS);
 18 }
 19
 20 static void             /* Handler for SIGTSTP */
 21 tstpHandler(int sig)
 22 {
 23     struct termios ourTermios;          /* To save our tty settings */
 24     sigset_t tstpMask, prevMask;
 25     struct sigaction sa;
 26     int savedErrno;
 27
 28     savedErrno = errno;                 /* We might change 'errno' here */
 29
 30     /* Save current terminal settings, restore terminal to
 31        state at time of program startup */
 32
 33     if (tcgetattr(STDIN_FILENO, &ourTermios) == -1)
 34         errExit("tcgetattr");
 35     if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &userTermios) == -1)
 36         errExit("tcsetattr");
 37
 38     /* Set the disposition of SIGTSTP to the default, raise the signal
 39        once more, and then unblock it so that we actually stop */
 40
 41     if (signal(SIGTSTP, SIG_DFL) == SIG_ERR)
 42         errExit("signal");
 43     raise(SIGTSTP);
 44
 45     sigemptyset(&tstpMask);
 46     sigaddset(&tstpMask, SIGTSTP);
 47     if (sigprocmask(SIG_UNBLOCK, &tstpMask, &prevMask) == -1)
 48         errExit("sigprocmask");
 49
 50     /* Execution resumes here after SIGCONT */
 51
 52     if (sigprocmask(SIG_SETMASK, &prevMask, NULL) == -1)
 53         errExit("sigprocmask");         /* Reblock SIGTSTP */
 54
 55     sigemptyset(&sa.sa_mask);           /* Reestablish handler */
 56     sa.sa_flags = SA_RESTART;
 57     sa.sa_handler = tstpHandler;
 58     if (sigaction(SIGTSTP, &sa, NULL) == -1)
 59         errExit("sigaction");
 60
 61     /* The user may have changed the terminal settings while we were
 62        stopped; save the settings so we can restore them later */
 63
 64     if (tcgetattr(STDIN_FILENO, &userTermios) == -1)
 65         errExit("tcgetattr");
 66
 67     /* Restore our terminal settings */
 68
 69     if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &ourTermios) == -1)
 70         errExit("tcsetattr");
 71
 72     errno = savedErrno;
 73 }
 74
 75 int
 76 main(int argc, char *argv[])
 77 {
 78     char ch;
 79     struct sigaction sa, prev;
 80     ssize_t n;
 81
 82     sigemptyset(&sa.sa_mask);
 83     sa.sa_flags = SA_RESTART;
 84
 85     if (argc > 1) {                     /* Use cbreak mode */
 86         if (ttySetCbreak(STDIN_FILENO, &userTermios) == -1)
 87             errExit("ttySetCbreak");
 88
 89         /* Terminal special characters can generate signals in cbreak
 90            mode. Catch them so that we can adjust the terminal mode.
 91            We establish handlers only if the signals are not being ignored. */
 92
 93         sa.sa_handler = handler;
 94
 95         if (sigaction(SIGQUIT, NULL, &prev) == -1)
 96             errExit("sigaction");
 97         if (prev.sa_handler != SIG_IGN)
 98             if (sigaction(SIGQUIT, &sa, NULL) == -1)
 99                 errExit("sigaction");
100
101         if (sigaction(SIGINT, NULL, &prev) == -1)
102             errExit("sigaction");
103         if (prev.sa_handler != SIG_IGN)
104             if (sigaction(SIGINT, &sa, NULL) == -1)
105                 errExit("sigaction");
106
107         sa.sa_handler = tstpHandler;
108
109         if (sigaction(SIGTSTP, NULL, &prev) == -1)
110             errExit("sigaction");
111         if (prev.sa_handler != SIG_IGN)
112             if (sigaction(SIGTSTP, &sa, NULL) == -1)
113                 errExit("sigaction");
114     } else {                            /* Use raw mode */
115         if (ttySetRaw(STDIN_FILENO, &userTermios) == -1)
116             errExit("ttySetRaw");
117     }
118
119     sa.sa_handler = handler;
120     if (sigaction(SIGTERM, &sa, NULL) == -1)
121         errExit("sigaction");
122
123     setbuf(stdout, NULL);               /* Disable stdout buffering */
124
125     for (;;) {                          /* Read and echo stdin */
126         n = read(STDIN_FILENO, &ch, 1);
127         if (n == -1) {
128             errMsg("read");
129             break;
130         }
131
132         if (n == 0)                     /* Can occur after terminal disconnect */
133             break;
134
135         if (isalpha((unsigned char) ch))        /* Letters --> lowercase */
136             putchar(tolower((unsigned char) ch));
137         else if (ch == '\n' || ch == '\r')
138             putchar(ch);
139         else if (iscntrl((unsigned char) ch))
140             printf("^%c", ch ^ 64);     /* Echo Control-A as ^A, etc. */
141         else
142             putchar('*');               /* All other chars as '*' */
143
144         if (ch == 'q')                  /* Quit loop */
145             break;
146     }
147
148     if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &userTermios) == -1)
149         errExit("tcsetattr");
150     exit(EXIT_SUCCESS);
151 }
